#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <inttypes.h>

int usage(char *argv0, char *msg)
{
	fprintf(stderr,
		"Copyright (C) 2014, Michael Gruhn <michael.gruhn@cs.fau.de>\n"
		"No rights reserved.\n"
		"\n"
		"usage: %s <ram.dump>\n"
		"\n"
		"Error: %s\n"
		, argv0, msg);
	return 1;
}


int die(char *msg)
{
	fprintf(stderr, "ERROR: %s\n", msg);
	return 1;
}


#include <sys/stat.h>

uintmax_t fsize(const char *filename)
{
	struct stat st;
	if( stat(filename, &st) )
		return 0;
	return st.st_size;
}


void* fmap(const char *filename)
{
	void *p;
	uintmax_t sz = fsize(filename);
	FILE *f = fopen(filename, "rb");
	if( f==NULL ) return NULL;
	p = malloc(sz);
	if(p)
	{
		if( fread(p,sz,1,f)!=1 )
		{
			free(p);
			p = NULL;
		}
	}
	
	fclose(f);
	return p;
}


int main(int argc, char *argv[])
{
	if(argc != 2)
		return usage(argv[0], "Argument count mismatch!");

	uint8_t *mem = fmap(argv[1]);

	printf("searching ...\n");

	uint64_t i;
	for(i=0; i<fsize(argv[1])-0x200;i+=0x8)
	{
		if( *(uint64_t*)&mem[i+0x18] ==  0 ||
		    *(uint64_t*)&mem[i+0x18] % 0x20 != 0 ||
		    *(uint32_t*)&mem[i+0x2c] <  0x80000000 ||
		    *(uint32_t*)&mem[i+0x30] <  0x80000000 ||
		    mem[i+0x00] != 0x03 ||
		    mem[i+0x01] != 0x00 ||
		    mem[i+0x02] != 0x26 ||
		    mem[i+0x03] != 0x00 ||
		    0)
			continue;
		printf("0x%08x: DirectoryTableBase=0x%08x,UniqueProcessID=%u,ImageFileName=%s\n", i, *(uint32_t*)&mem[i+0x18], *(uint32_t*)&mem[i+0xb4], &mem[i+0x16c]);
#if 0
		printf("0x%08x 0x%08x, 0x%"PRIx64", 0x%08x, 0x%08x, %x, %x, %x, %x\n",
		*(uint32_t*)&mem[i+0x18],
		*(uint32_t*)&mem[i+0x1c],
		*(uint64_t*)&mem[i+0x18],
		    *(uint32_t*)&mem[i+0x2c],
		    *(uint32_t*)&mem[i+0x30],
		    mem[i+0x00],
		    mem[i+0x01],
		    mem[i+0x02],
		    mem[i+0x03]);
		//break;
#endif
	}

	free(mem);

	return 0;
}


