#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <inttypes.h>


int usage(char *argv0, char *msg)
{
	fprintf(stderr,
		"Copyright (C) 2014, Michael Gruhn <michael.gruhn@cs.fau.de>\n"
		"No rights reserved.\n"
		"\n"
		"usage: %s <ram.dump>\n"
		"\n"
		"Error: %s\n"
		, argv0, msg);
	return 1;
}


int die(char *msg)
{
	fprintf(stderr, "ERROR: %s\n", msg);
	return 1;
}


#include <sys/stat.h>

uintmax_t fsize(const char *filename)
{
	struct stat st;
	if( stat(filename, &st) )
		return 0;
	return st.st_size;
}


void* fmap(const char *filename)
{
	void *p;
	uintmax_t sz = fsize(filename);
	FILE *f = fopen(filename, "rb");
	if( f==NULL ) return NULL;
	p = malloc(sz);
	if(p)
	{
		if( fread(p,sz,1,f)!=1 )
		{
			free(p);
			p = NULL;
		}
	}
	
	fclose(f);
	return p;
}


typedef struct EPROCESS
{
	uint8_t  Type; //   +0x000 Type             : UChar
	uint8_t  Flags; //     +0x001 Signalling       : UChar
	uint8_t  Size; //    +0x002 Size             : UChar
	uint8_t  Reserved1; //   +0x003 Reserved1        : UChar
	uint8_t  offset0[0x14];
	uint32_t DirectoryTableBase; //      +0x018 DirectoryTableBase : Uint4B
	uint8_t  offset1[0x10];
	uint32_t ThreadListHead_Flink;//   +0x02c ThreadListHead   : _LIST_ENTRY
	uint32_t ThreadListHead_Blink;//
	uint8_t  offset2[0x80];
	uint32_t UniqueProcessId; //     +0x0b4 UniqueProcessId  : Ptr32 Void
	uint32_t ActiveProcessLinks_Flink; //     +0x0b8 ActiveProcessLinks : _LIST_ENTRY
	uint32_t ActiveProcessLinks_Blink;
	uint8_t  offset3[0xb0];
	uint8_t  ImageFileName[0x15]; //   +0x170 ImageFileName    : [15] UChar
} EPROCESS __attribute__((packed));


#include <ctype.h>

static int isImageFileName(char *name)
{
	unsigned i=0;
	for(i=0;i<15;i++)
	{
		if( name[i] == '\0' ) break;
		if( !isprint(name[i]) ) return 0;
	}
	if( i == 15 ) return 0;
	return 1;
}


static void match(uint64_t addr, void *ptr)
{
	EPROCESS *p = (EPROCESS*)ptr;

	if(
		p->ThreadListHead_Flink >= 0x80000000 &&
		p->ThreadListHead_Blink >= 0x80000000 &&
		p->DirectoryTableBase != 0 &&
		p->DirectoryTableBase % 0x20 == 0 &&
		p->Type == 0x03 &&
		p->Flags == 0x00 &&
		p->Size == 0x28 &&
		p->Reserved1 == 0x00 &&
		isImageFileName(p->ImageFileName) &&
	1 )
	{
		printf("0x%016"PRIx64": DirectoryTableBase=0x%016"PRIx64",UniqueProcessId=%"PRIu64",ImageFileName=%s\n",
			addr, p->DirectoryTableBase, p->UniqueProcessId, p->ImageFileName);
	#if 1
		printf("0x%016"PRIx64": DirectoryTableBase=0x%016"PRIx64",UniqueProcessId=%"PRIu64",ImageFileName=%s, %x, %x, %x, %x\n",
			addr, p->DirectoryTableBase, p->UniqueProcessId, p->ImageFileName, p->Type, p->Flags, p->Size, p->Reserved1);
	#endif
	}
}


int main(int argc, char *argv[])
{
	if(argc != 2) return usage(argv[0], "Argument count mismatch!");

	uint8_t *mem = fmap(argv[1]);

	printf("searching ...\n");

	uint64_t i;
	for(i=0; i<fsize(argv[1])-0x200;i+=0x8) match(i, &mem[i]);

	free(mem);

	return 0;
}


