#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>


int usage(char *argv0, char *msg)
{
	fprintf(stderr,
		"Copyright (C) 2014, Michael Gruhn <michael.gruhn@cs.fau.de>\n"
		"No rights reserved.\n"
		"\n"
		"usage: %s <size> <alloc> <fill>\n"
		"\tsize\tsize in MiB to allocate\n"
		"\talloc\talloc mode to use, supported modes are:\n"
		"\t\tmalloc\tuse malloc()\n"
		"\t\tcalloc\tuse calloc()\n"
		"\tfill\tfill mode to use, supported modes are:\n"
		"\t\tno\tonly allocate but not fill\n"
		"\t\tzero\tfill with zero bits\n"
		"\t\tone\tfill with one bits\n"
		"\t\taddr\tfill with 32 bit address\n"
		"\t\tdeadbeef\tfill with 0xdeadbeef\n"
		"\t\trandom\tfill with random 32 bit numbers\n"
		"\n"
		"Error: %s\n"
		, argv0, msg);
	return 1;
}


int die(char *msg)
{
	fprintf(stderr, "ERROR: %s\n", msg);
	return 1;
}


int main(int argc, char *argv[])
{
	if(argc != 4)
		return usage(argv[0], "Argument count mismatch!");

	size_t sz = strtol(argv[1],NULL,10)*1024*1024/4;
	uint32_t *unalignedmem = NULL;
	uint32_t *mem = NULL;

	if(strcmp(argv[2],"malloc")==0)
	{
		unalignedmem = malloc((sz*sizeof(*mem))+0x1000);
		if(unalignedmem==NULL)
			return die(strerror(errno));
		size_t j;
		for(j=0; j<0x1000/4; j++)
			unalignedmem[j] = 0xc001babe;
	}
	else if(strcmp(argv[2],"calloc")==0)
	{
		unalignedmem = calloc((sz*sizeof(*mem))+0x1000, 1);
		if(unalignedmem==NULL)
			return die(strerror(errno));
	}
	else
	{
		return usage(argv[0], "Invalid allocation mode!");
	}

	mem = (void*)((((uintptr_t)unalignedmem+0xfff)/0x1000)*0x1000);

	size_t i;
	if(strcmp(argv[3],"no")==0)
	{
	}
	else if(strcmp(argv[3],"zero")==0)
	{
		for(i=0; i<sz; i++)
			mem[i] = 0x00000000;
	}
	else if(strcmp(argv[3],"one")==0)
	{
		for(i=0; i<sz; i++)
			mem[i] = 0xffffffff;
	}
	else if(strcmp(argv[3],"deadbeef")==0)
	{
		for(i=0; i<sz; i++)
			mem[i] = 0xdeadbeef;
	}
	else if(strcmp(argv[3],"random")==0)
	{
		uint32_t x = 0xdeadbeef;
		for(i=0; i<sz; i++)
			x = mem[i] = 1664525*x + 1013904223;
	}
	else if(strcmp(argv[3],"addr")==0)
	{
		for(i=0; i<sz; i++)
			mem[i] = (uint32_t)i;
	}
	else
	{
		return usage(argv[0], "Invalid fill mode!");
	}
	printf("Success: 0x%08x\n", mem);
	getchar();
	return 0;
}


