#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <inttypes.h>

int usage(char *argv0, char *msg)
{
	fprintf(stderr,
		"Copyright (C) 2014, Michael Gruhn <michael.gruhn@cs.fau.de>\n"
		"No rights reserved.\n"
		"\n"
		"usage: %s <phymem.dd> <DirectoryTableBase> <pagefile.sys> <virmem.dd> <virmem.map>\n"
		"\n"
		"Error: %s\n"
		, argv0, msg);
	return 1;
}


int die(char *msg)
{
	fprintf(stderr, "ERROR: %s\n", msg);
	return 1;
}


#include <sys/stat.h>
#include <assert.h>

uintmax_t fsize(const char *filename)
{
	struct stat st;
	assert( stat(filename, &st)==0 );
	return st.st_size;
}


void* fmap(const char *filename)
{
	void *p;
	uintmax_t sz = fsize(filename);
	FILE *f = fopen(filename, "rb");
	if( f==NULL ) return NULL;
	p = malloc(sz);
	if(p && fread(p,sz,1,f)!=1 )
	{
		free(p);
		p = NULL;
	}
	fclose(f);
	return p;
}


typedef struct virma_t
{
	void*    phymem;
	uint64_t phymem_sz;
	void*    pagefile;
	uint64_t pagefile_sz;
	void*    zeromem;
	uint64_t dtb;
} virma_t;


int virma_isswap(virma_t *this, void *p)
{
	if( p >= this->pagefile && p < this->pagefile+this->pagefile_sz ) return 1;
	return 0;
}


virma_t* virma_init(const char *phymempath, uint64_t dtb, const char *pagefilepath)
{
	fprintf(stderr,"virma_init(phymempath=%s,pagefilepath=%s,dtb=0x%x\n",phymempath,pagefilepath,dtb);

	virma_t *this = malloc(sizeof(*this));
fprintf(stderr,"1000\n");
	if(this==NULL) goto fail_malloc;

	this->zeromem = calloc(0x1000,1);
fprintf(stderr,"2000\n");
	if(this->zeromem==NULL) goto fail_calloc;

	this->phymem_sz = fsize(phymempath);
fprintf(stderr,"3000\n");
	this->phymem = fmap(phymempath);
fprintf(stderr,"4000\n");
	if(this->phymem==NULL) goto fail_fmap_phymem;

	this->pagefile_sz = fsize(pagefilepath);
fprintf(stderr,"5000\n");
	this->pagefile = fmap(pagefilepath);
fprintf(stderr,"6000\n");
	if(this->pagefile==NULL) goto fail_fmap_pagefile;
fprintf(stderr,"7000\n");

	this->dtb = dtb;

	return this;

fail_fmap_pagefile:
	free(this->phymem);
fail_fmap_phymem:
	free(this->zeromem);
fail_calloc:
	free(this);
fail_malloc:
	return this=NULL;
}



int fprintf_pe(FILE *f, const char *prefix, uint64_t pe)
{
	return fprintf(f,"%s0x%016"PRIx64" [t=%d,p=%d,i=0x%08x,ps=%d,d=%d,n=%d,v=%d]\n", prefix?prefix:"", pe,
		pe>>11&1, pe>>10&1,  (pe>>3&~0x3f) | (pe>>1&0x3f) , pe>>7&1, pe>>6&1, pe>>1&0xf, pe&1
	);
}

#define PS(pe)	((pe)>>7&1)
#define T(pe)	((pe)>>11&1)
#define P(pe)	((pe)>>10&1)
#define V(pe)	((pe)&1)
#define PF(pe)	((pe)>>1&0xf)

void* virma_vtop(virma_t *this, uint64_t v)
{
	fprintf(stderr,"virma_vtop(this=0x%p,v=0x%016"PRIx64")\n", this, v);
	uint64_t *t = this->phymem;
	uint64_t *tp = this->pagefile;
	uint64_t pdpte = t[   (this->dtb & 0x00000000ffffffe0 + (v>>30&0x3)*8)/8   ];
	uint64_t pde = 0;
	uint64_t pte = 0;
	uint64_t p = 0;

	int pf = 0;

	fprintf_pe(stderr,"pdpte:  ", pdpte);

	if( (pdpte&0xfffffffffffffc01)==0 ) // demand zeros
		return this->zeromem;
	else if( PS(pdpte) ) // 1 GiB frame
	{
		if( !V(pdpte) ) // not present
			return NULL;
		fprintf(stderr,"ERROR: 1 GiB PAE frame encountered."
		               "This is not implemented, "
		               "because it should not happen."
		               "NO REALLY 1 GiB PAE FRAMES DO NOT EXIST!");
		exit(1);
	}
	else if( !V(pdpte) )
	{
		puts("PDPTE not present!");
		exit(1);
		return NULL;
	}

	pde = t[(   (pdpte&0x000ffffffffff000) + (v>>21&0x1ff)*8   )/8];
	fprintf(stderr,"pde:    0x%016"PRIx64"\n", pde);
	fprintf_pe(stderr,"pde:    ", pde);

	if( (pde&0xfffffffffffffc01)==0 ) // demand zeros
		return this->zeromem;
	else if( !V(pde) && !P(pde) && !T(pde) ) // pagefile
	{
		puts("PTE pagefile");
		if( PF(pde) )
			exit(1);
		pte = tp[(   (pde>>20&0x00ffffffffff000) + (v>>12&0x1ff)*8   )/8];
		fprintf_pe(stderr,"pte:    ", pte);
		//exit(1);
	}
	else if( V(pde) && PS(pde) ) // 2 MiB frame ... never encountered
	{
		fprintf(stderr,"ERROR: 2 MiB frame encountered."
		               "This is not implemented, "
		               "because it should not happen.");
		exit(1);
	}
	else if( V(pde) || (T(pde) && !P(pde)) ) // present or transient
	{
		pte = t[(   (pde&0x000ffffffffff000) + (v>>12&0x1ff)*8   )/8];
		fprintf_pe(stderr,"pte:    ", pte);
	}
	else
	{
		puts("FOO?!?");
		exit(1);
		return NULL;
	}

	if( (pte&0xfffffffffffffc01)==0 ) // demand zeros
		return this->zeromem;
	else if( !V(pte) && !P(pte) && !T(pte) ) // pagefile
	{
		puts("pagefile");
		if( PF(pte) )
			exit(1);
		p = (pte>>20&0x000ffffffffff000) + (v&0xfff);
		fprintf(stderr,"p:       0x%016"PRIx64" (pagefile)\n", p);
		return (void*)&((uint8_t*)this->pagefile)[p];
	}
	else if( V(pte) || (T(pte) && !P(pte)) ) // present or transient
	{
	}
	else
	{
		puts("BAR?!?");
		if( T(pte) && P(pte) )
			exit(1);
		return NULL;
	}

	p = (pte&0x000ffffffffff000)+(v&0xfff);
	fprintf(stderr,"p:      0x%016"PRIx64"\n", p);

	return (void*)&((uint8_t*)this->phymem)[p];
}


void virma_fini(virma_t *this)
{
	free(this->phymem);
	free(this);
}


int main(int argc, char *argv[])
{
	if(argc != 6)
		return usage(argv[0], "Argument count mismatch!");
	
	virma_t *virma = virma_init(argv[1], strtol(argv[2],NULL,16), argv[3]);
	
	FILE *fvirmem = fopen(argv[4],"wb");
	FILE *fvirmap = fopen(argv[5],"wb");
	
	uint64_t v, o;
	
	fprintf(fvirmap, "Virtual Address           Physical Address          File Offset               Source\n");
	fprintf(fvirmap, "-+----------------        -+----------------        -+----------------        -+----\n");
	
//#define TEST
#ifndef TEST
	for(v=0x0,o=0; v<0x80000000; v+=0x1000)
#else
	fvirmap = stdout;
//	v = 0x00ab1000;
//	v = 0x00ac1000;
//	v = 0x005f1000;
//	v = 0x00771000;
//	v = 0x1000000;
//	v = 0x9ff000;
//	v = 0x00806000;
//	v = 0x0009f0000;
//	v = 0x001000000;
//	v = 0x0010b1000;
	o = 0;
#endif
	{
		void *p = virma_vtop(virma, v);
#ifndef TEST
		if(p==NULL) continue;
		if(p==virma->zeromem) continue;
#else
		if(p==NULL) p==virma->zeromem;
#endif
		fprintf(fvirmap,"0x%016"PRIx64"        0x%016"PRIx64"        0x%016"PRIx64"        %s\n",
			v,
			((uint64_t)p) - ((uint64_t)(virma_isswap(virma, p)?virma->pagefile:virma->phymem)),
			o,
			virma_isswap(virma, p)?"pagefile.sys":"RAM"
		);
		o+=0x1000;
		printf("%08"PRIx32" %08"PRIx32"\n", *(uint32_t*)p, *(uint32_t*)(p+4));
		fwrite(p, 0x1000, 1, fvirmem);
	}
	
	fclose(fvirmem);
	fclose(fvirmap);
	virma_fini(virma);
	
	return 0;
}


